/* 
 * File:   MED_mesh.h
 * Author: Andriy Andreykiv, Delft, The Netherlands
 *         andriy.andreykiv@gmail.com
 * * Created on July 21, 2008, 2:31 PM
 */

#ifndef _MED_MESH_H
#define	_MED_MESH_H

#include "MEDMEM_Mesh.hxx"
#include "MEDMEM_Group.hxx"
#include "MEDMEM_Field.hxx"

#include <getfem/getfem_mesh.h>
#include <getfem/getfem_mesh_fem.h>
#include <map>
#include <vector>
#include <ext/hash_map>
#include <math.h>
#include <getfem/getfem_modeling.h>



typedef getfem::modeling_standard_plain_vector  plain_vector;
using namespace MEDMEM;
using namespace MED_EN;
using namespace DRIVERFACTORY;
using bgeot::size_type; /* = unsigned long */
using bgeot::base_node; /* geometrical nodes (derived from base_small_vector)*/
inline size_t ipower(int x,  int n){int pow=1; for(int i=0;i<=n;i++) pow*=x; return size_t(pow);} //power of an integer number
struct face{size_type cv,f;}; //structure for locating an element face or an edge
struct compare_vectors{ // algorithm class for comparing two vectors (used for hashing element faces)
    bool operator()(const std::vector<size_type> v1, const std::vector<size_type> v2) const
        {
            for(size_type i=0;i<v1.size();i++)
                if(v1[i]!=v2[i]) return false;
        return true;
        }
};
struct hash_function{ //algorithm class for hash function
    size_t operator()(const std::vector<size_type> V) const
    {
        size_t sum=0;
        for(size_t i=0;i<V.size();i++) sum+=size_t(V[i]*ipower(10,i));
        return sum;
    }

};

// a interface to SALOME MED class with 
// capability to read node, face, edge and elements sets
// and export nodal fields
class MED_mesh{
    MESH imported_med_mesh; // imported MED MESH
    getfem::mesh gfem_mesh; // a get_fem mesh generated from the imported MED mesh
    static std::map<medGeometryElement, std::string> med_to_getfem_geotype;  // a table that converts MED to getfem element types
    static bool types_initialized;             // a state variable for the above
    std::vector<size_type> med_getfem_node;    // MED to getfem node numbering
    std::vector<size_type> med_getfem_element; // MED to getfem element numbering
    std::map<std::string,size_type> name_gfem_group;  //MED element group name to getfem region number
    std::map<std::string,int> name_node_group;        //MED nodal group name to getfem nodal group number
    std::vector<dal::bit_vector> node_group;          //a container for getfem nodal groups
    __gnu_cxx::hash_map<std::vector<size_type>,face,hash_function,compare_vectors> connect_to_face; //maps the face nodes to the getfem face nodes
    void build_getfem_mesh(void);
    void build_getfem_regions(void);
    void initialize_types(void);

public:

    MED_mesh(std::string fileName, std::string mesh_name) :  // provide MED file name and the name of the MESH inside it
    imported_med_mesh(MED_DRIVER, fileName, mesh_name) {
        std::cout<<"MED_mesh: importing mesh "<<mesh_name<<" from "<<fileName<<std::endl;
        if (!types_initialized) initialize_types();
        build_getfem_mesh();
        build_getfem_regions();
        std::cout<<"MED_mesh: "<<fileName<<" loaded into memory"<<std::endl;
    }
     MED_mesh(MESH copy_from_med_mesh) : imported_med_mesh(copy_from_med_mesh)
    {
        if (!types_initialized) initialize_types();
        build_getfem_mesh();
        build_getfem_regions();
    }

    getfem::mesh& get_getfem_mesh() {
        return gfem_mesh;
    }
    size_type get_region_index(std::string group_name){return name_gfem_group[group_name];} //returns getfem region index from MED element (face) group  name
    int get_node_group_index(std::string name){return name_node_group[name];}               //returns getfem nodal group index from MED nodal group name
    dal::bit_vector get_node_group(int node_group_number){return node_group[node_group_number];} //returns nodal group by its index
    dal::bit_vector get_node_group(std::string name){return node_group[name_node_group[name]];}  //returns nodal group by its MED nodal group name
    int nb_nodal_groups(void){return node_group.size();}                                          //number of nodal groups
    void write_nodal_field(getfem::mesh_fem& mf, plain_vector U, int inc,double  time, std::string field_name, std::string file_name);  //writes a nodal field from a getfem result vector
};
#endif	/* _MED_MESH_H */


















