#include "MED_mesh.h"
#include "itoa.h"
#include <algorithm>

std::map<medGeometryElement, std::string> MED_mesh::med_to_getfem_geotype;
bool MED_mesh::types_initialized=false;

void MED_mesh::initialize_types() { //building correspondence between MED and getfem element types
        med_to_getfem_geotype[MED_SEG2] =  "GT_PK(1,1)";
        med_to_getfem_geotype[MED_SEG3] =  "GT_PK(1,2)";
        med_to_getfem_geotype[MED_TRIA3] = "GT_PK(2,1)";
        med_to_getfem_geotype[MED_TRIA6] = "GT_PK(2,2)";
        med_to_getfem_geotype[MED_QUAD4] = "GT_QK(2,1)"; 
        med_to_getfem_geotype[MED_QUAD8] = "GT_QK(2,2)";  //this is probably wrong !!! need to elaborate
        med_to_getfem_geotype[MED_TETRA4] ="GT_PK(3,1)";
        med_to_getfem_geotype[MED_TETRA10]="GT_PK(3,2)";  
        med_to_getfem_geotype[MED_HEXA8] = "GT_QK(3,1)";
        med_to_getfem_geotype[MED_HEXA20] ="GT_QK(3,2)";  //this is also wrong !!! need to check
        med_to_getfem_geotype[MED_PENTA6] ="GT_PRISM(3,1)"; //to be checked
        types_initialized = true; 
}

void MED_mesh::build_getfem_mesh() {
    //importing nodes from the MED file
    int SpaceDimension = imported_med_mesh.getSpaceDimension() ;
    int NumberOfNodes =  imported_med_mesh.getNumberOfNodes() ;
    const double * Coordinates = imported_med_mesh.getCoordinates(MED_FULL_INTERLACE) ; 
    med_getfem_node.resize(NumberOfNodes+1);
    med_getfem_node[0]=size_type(0);
    for(int i=0; i<NumberOfNodes ; i++) {
        bgeot::base_node node;
        switch (SpaceDimension){
            case 1:
                {
                    med_getfem_node[i+1]=gfem_mesh.add_point(bgeot::base_node(Coordinates[i]));
                    break;
                }
            case 2:
                {
                    med_getfem_node[i+1]=gfem_mesh.add_point(bgeot::base_node(Coordinates[2*i],Coordinates[2*i+1])); 
                    break;
                }
            case 3:
                {
                    med_getfem_node[i+1]=gfem_mesh.add_point(bgeot::base_node(Coordinates[3*i],Coordinates[3*i+1],Coordinates[3*i+2]));
                }          
        }
     }
    // reading connectivity and creating the elements
    int number_of_types = imported_med_mesh.getNumberOfTypes(MED_CELL);
    const medGeometryElement * Types = imported_med_mesh.getTypes(MED_CELL);
    const int * connectivity_index = imported_med_mesh.getConnectivityIndex(MED_NODAL, MED_CELL);
    const int * connectivity = imported_med_mesh.getConnectivity(MED_FULL_INTERLACE, MED_NODAL, MED_CELL, MED_ALL_ELEMENTS);
    const int * element_index= imported_med_mesh.getGlobalNumberingIndex(MED_CELL);
    int number_of_elements_in_mesh=imported_med_mesh.getNumberOfElements(MED_CELL,MED_ALL_ELEMENTS);
    med_getfem_element.resize(number_of_elements_in_mesh+1);
    med_getfem_element[0]=size_type(0);
    //iteration over every single element type
    for (int i = 0; i < number_of_types; i++) {
        medGeometryElement myType = Types[i];
        bgeot::pgeometric_trans pgt = bgeot::geometric_trans_descriptor(med_to_getfem_geotype[myType]);

        // Connectivity retrieval
        for (int j = element_index[i]; j < element_index[i+1]; j++) {
            std::vector<size_type> element_nodes;
            for (int k = connectivity_index[j-1]; k < connectivity_index[j]; k++)
                   element_nodes.push_back(med_getfem_node[connectivity[k - 1]]);
            
            std::vector<size_type> elm_nodes_gfem(element_nodes.size());
            switch (myType) {   // this is an element specific conversion that accounts for difference 
                                // in node numbering between Getfem and MED
                case MED_QUAD4:
                {
                    elm_nodes_gfem[0] = element_nodes[0];
                    elm_nodes_gfem[1] = element_nodes[1];
                    elm_nodes_gfem[2] = element_nodes[3];
                    elm_nodes_gfem[3] = element_nodes[2];
                    break;
                }
                case MED_TRIA6:
                {
                    elm_nodes_gfem[0] = element_nodes[0];
                    elm_nodes_gfem[1] = element_nodes[5];
                    elm_nodes_gfem[2] = element_nodes[2];
                    elm_nodes_gfem[3] = element_nodes[3];
                    elm_nodes_gfem[4] = element_nodes[4];
                    elm_nodes_gfem[5] = element_nodes[1];
                    break;
                }
                case MED_SEG3:
                {   
                    elm_nodes_gfem[0] = element_nodes[0];
                    elm_nodes_gfem[1] = element_nodes[2];
                    elm_nodes_gfem[2] = element_nodes[1];
                    break;
                }
                default:
                    gmm::copy(element_nodes, elm_nodes_gfem);  // here we blindly copy MED node elements to Getfem
		                                               // elements disregarding the difference in numbering
		                                               // this has to be checked for all elements and edited if needed
            }

            med_getfem_element[j]=gfem_mesh.add_convex(pgt,elm_nodes_gfem.begin());
        }
    }
    
    //Here we are hashing the nodes of the faces/edge elements of the MED mesh, to be able
    //to map them to the edge/faces of Getfem mesh
    getfem::mesh_region outer_faces;
    getfem::outer_faces_of_mesh(gfem_mesh, outer_faces);
    for(getfem::mr_visitor i(outer_faces); !i.finished();++i)
    {
        std::vector<size_type> face_nodes;
        bgeot::pconvex_structure cvs =  gfem_mesh.structure_of_convex(i.cv());
        for(size_type k=0;k<cvs->nb_points_of_face(i.f());k++) 
	    face_nodes.push_back(gfem_mesh.ind_points_of_face_of_convex(i.cv(),i.f())[k]);  
        std::sort(face_nodes.begin(),face_nodes.end());
        face this_face;  
        this_face.cv=i.cv();
        this_face.f=i.f();
        connect_to_face[face_nodes]=this_face; 
    }
}


void MED_mesh::build_getfem_regions()
{ 
    // IMPORTING ELEMENT GROUPS
    int number_of_element_groups=imported_med_mesh.getNumberOfGroups(MED_CELL); 
    const vector<GROUP*> element_group=imported_med_mesh.getGroups(MED_CELL);
    for(int i=0;i<number_of_element_groups;i++)
    {  
        GROUP * pGroup=element_group[i];
        name_gfem_group[pGroup->getName()]=i+1;
        std::cout<<"Importing element group "<<pGroup->getName()<<" into getfem_mesh region "<<i+1<<std::endl;
        if(pGroup->getName()=="Group_Of_All_Faces" || 
           pGroup->getName()=="Group_Of_All_Edges" || 
           pGroup->getName()=="Group_Of_All_Volumes")
        {
            gfem_mesh.region(i+1).add(gfem_mesh.convex_index());
            continue;
        }
        const int number_of_elements_group=pGroup->getNumberOfElements(MED_ALL_ELEMENTS);
        std::cout<<"number of elements in this group "<<number_of_elements_group<<std::endl;
        const int* group_element=pGroup->getNumber(MED_ALL_ELEMENTS); 
        std::cout<<"got the group index"<<std::endl;
        for(int j=0;j<number_of_elements_group;j++)
        {   std::cout<<"       imported med elements: "<<group_element[j]<<" getfem element: "<<med_getfem_element[group_element[j]]<<std::endl;;
            gfem_mesh.region(i+1).add(med_getfem_element[group_element[j]]); 
        }
    }
    
    // IMPORTING BOUNDARY GROUPS (FACES AND EDGES)
    // reading connectivity and creating the elements
    medGeometryElement boundary_entity;
    int MeshDimension = imported_med_mesh.getMeshDimension();
    boundary_entity=(MeshDimension==2) ? MED_EDGE : MED_FACE;
    const vector<GROUP*> face_group=imported_med_mesh.getGroups(boundary_entity);
    int number_of_face_groups=imported_med_mesh.getNumberOfGroups(boundary_entity);
    if(MeshDimension==1) number_of_face_groups=0;      
    const int * connectivity_index = imported_med_mesh.getConnectivityIndex(MED_NODAL, boundary_entity);
    const int * connectivity = imported_med_mesh.getConnectivity(MED_FULL_INTERLACE, MED_NODAL, boundary_entity, MED_ALL_ELEMENTS);
    for(int i=0;i<number_of_face_groups;i++)
    {
         GROUP * pGroup=face_group[i];
         name_gfem_group[pGroup->getName()]=i+number_of_element_groups+1;       
         std::cout<<"Importing face group "<<pGroup->getName()<<" into getfem_mesh region "<<i+1+number_of_element_groups<<std::endl;
         if(pGroup->getName()=="Group_Of_All_Faces" || 
           pGroup->getName()=="Group_Of_All_Edges" )
         {  
            getfem::mesh_region outer_faces;
            getfem::outer_faces_of_mesh(gfem_mesh, outer_faces);
            gfem_mesh.region(i+1).add(outer_faces.index());
            continue;
         }
         const int* group_element=pGroup->getNumber(MED_ALL_ELEMENTS);
         const int number_of_elements_group=pGroup->getNumberOfElements(MED_ALL_ELEMENTS);
         for(int j=0;j<number_of_elements_group;j++)
         {  
            // Connectivity retrieval of face j
            std::vector<size_type> gface_nodes;
            for (int k = connectivity_index[group_element[j]-1]; k < connectivity_index[group_element[j]]; k++) 
                gface_nodes.push_back(med_getfem_node[connectivity[k - 1]]);
            std::sort(gface_nodes.begin(),gface_nodes.end());
            // using face_nodes obtain iterator to the element and face number
            face gfem_face=connect_to_face[gface_nodes]; // using the hashed table
            // add this face to the gfem_mesh region 
            gfem_mesh.region(i+number_of_element_groups+1).add(gfem_face.cv,gfem_face.f);  
              //std::cout<<"Group element "<<j<<" adding face "<<gfem_face.f;
              //std::cout<<" of element "<<gfem_face.cv;
              //std::cout<<" to gfem region "<<name_gfem_group[pGroup->getName()];
              //std::cout<<" of MED boundary group "<<pGroup->getName()<<std::endl;
            
         } // finished the current group
         
    } //finished all boundary groups

    
    // IMPORTING NODE GROUPS
    int number_of_node_groups=imported_med_mesh.getNumberOfGroups(MED_NODE); 
    const vector<GROUP*> med_node_group=imported_med_mesh.getGroups(MED_NODE);
    for(int i=0;i<number_of_node_groups;i++)
    {  
        GROUP * pGroup=med_node_group[i];
        name_node_group[pGroup->getName()]=i;
        std::cout<<"Importing node group "<<pGroup->getName()<<" into vector number "<<i<<std::endl;
        if(pGroup->getName()=="Group_Of_All_Nodes")
        {
            node_group.push_back(gfem_mesh.points_index());
            continue;
        }
        const int number_of_nodes_in_group=pGroup->getNumberOfElements(MED_ALL_ELEMENTS);
        std::cout<<"number of nodes in this group "<<number_of_nodes_in_group<<std::endl;
        const int* group_element=pGroup->getNumber(MED_ALL_ELEMENTS); 
        dal::bit_vector current_group;
        for(int j=0;j<number_of_nodes_in_group;j++)
            current_group.add(med_getfem_node[group_element[j]]);

        node_group.push_back(current_group);
    } // finished node groups
}

void MED_mesh::write_nodal_field(getfem::mesh_fem& mf_u, plain_vector U,int inc, double time, std::string field_name, std::string file_name)
{
        SUPPORT AllNodesSupport(&imported_med_mesh,"Support on all Nodes",MED_NODE);
        int NumberOfComponents=mf_u.get_qdim();
        FIELD<double> nodal_field(&AllNodesSupport,NumberOfComponents);
        nodal_field.setName(field_name);  
        string * ComponentsNames = new string[NumberOfComponents] ;
        for(int i=0;i<NumberOfComponents; i++) ComponentsNames[i]="component "+itoa(i+1,10);
        nodal_field.setComponentsNames(ComponentsNames) ;
        string * ComponentsDescriptions = new string[NumberOfComponents] ;
        for(int i=0;i<NumberOfComponents; i++) ComponentsDescriptions[i] = "magnitude  "+itoa(i+1,10)+" of some stuff";
        nodal_field.setComponentsDescriptions(ComponentsDescriptions) ;
        string * ComponentsUnits = new string[NumberOfComponents] ;
        for(int i=0;i<NumberOfComponents; i++) ComponentsUnits[i] = "$";  
        nodal_field.setMEDComponentsUnits(ComponentsUnits) ;
        nodal_field.setIterationNumber(inc);
        nodal_field.setOrderNumber(MED_NONOR);  
        nodal_field.setTime(time);
        
  
           //building the DOF correspondence index
        __gnu_cxx::hash_map<size_type, std::vector<size_type> > NODE_DOF;
        //std::cout << "   Buildig the DOF correspondence index" << std::endl;
        for (dal::bv_visitor e(mf_u.convex_index()); !e.finished(); ++e) {
            std::vector<size_type> nodes_e = mf_u.linked_mesh().ind_points_of_convex(e);
            size_type dof_e = 0;
            for (size_type node = size_type(0); node < nodes_e.size(); node++) {
               std::vector<size_type> dofs(mf_u.get_qdim());
               for (size_type i = 0; i < mf_u.get_qdim(); i++, dof_e++) dofs[i] = mf_u.ind_dof_of_element(e)[dof_e];
               NODE_DOF[nodes_e[node]] = dofs;
            }
        }
        //std::cout<<"   .... done"<<std::endl;
        
        //setting the field values
        // Value :
        int NumberOfValue = AllNodesSupport.getNumberOfElements(MED_ALL_ELEMENTS);
        for(int i=1; i<=NumberOfValue; i++) // i^th element
            for (int j=1; j<=NumberOfComponents; j++) { // j^th component
                size_type gfem_dof=NODE_DOF[med_getfem_node[i]][j-1];
                double myValue = U[gfem_dof];
                nodal_field.setValueIJ(i,j,myValue);
            }
        int id_mesh=imported_med_mesh.addDriver(MED_DRIVER,file_name,imported_med_mesh.getName());
        imported_med_mesh.write(id_mesh);
        int id_field = nodal_field.addDriver(MED_DRIVER,file_name,field_name);
        nodal_field.write(id_field);
}






















