#include <getfem/getfem_model_solvers.h>
#include <getfem/getfem_modeling.h>
#include <gmm/gmm.h>
#include <getfem/getfem_assembling.h>
#include <getfem/getfem_regular_meshes.h>
#include <getfem/getfem_mesh.h>
#include <getfem/getfem_mesh_fem.h>
#include <getfem/getfem_export.h>
#include <getfem/dal_bit_vector.h>
#include <getfem/getfem_nonlinear_elasticity.h>
#include <math.h>
#include <fstream>
#include "MED_mesh.h" 

/* some Getfem++ types that we will be using  */
using bgeot::base_small_vector; /* special class for small (dim<16) vectors */
using bgeot::base_node;  /* geometrical nodes(derived from base_small_vector)*/
using bgeot::scalar_type; /* = double */
using bgeot::size_type;   /* = unsigned long */
using bgeot::base_matrix; /* small dense matrix. */
using bgeot::base_vector;

/* definition of some matrix/vector types. 
 * default types of getfem_model_solvers.h
 */
typedef getfem::modeling_standard_sparse_vector sparse_vector;
typedef getfem::modeling_standard_sparse_matrix sparse_matrix;
typedef getfem::modeling_standard_plain_vector  plain_vector;

std::string itoa(int value, int base);
getfem::mesh* deform_mesh(const getfem::mesh_fem& mf,plain_vector& def_U);
void export_vtk(const getfem::mesh_fem&,const std::vector<bgeot::scalar_type>&, std::string);
void copy_regions(const getfem::mesh& from_mesh, getfem::mesh& to_mesh);
struct DVtable{ scalar_type dispY, volt;};
struct finite_strain_problem{
	ftool::md_param PARAM;
	std::vector<DVtable> dvtab;
	base_vector mat_implant_param;
	bool read_flag;
	enum { NORTH = 1, EAST = 2, WEST = 3, SOUTH = 4, EA_WE=5, BOUNDARY=6};
        MED_mesh med_mechanic;
	getfem::mesh m_mechanic;		    //initial mechanic mesh
	getfem::mesh_fem mf_mechanic;
	getfem::mesh_im  mim_mechanic;
	getfem::mesh_fem mf_data;
	getfem::mdbrick_source_term<>* pstruct_brick;
	
	scalar_type TOLERANCE_NEWTON, MAT_LAMBDA, MAT_MU,P_FINAL,MAX_ITER;
	std::string MECHA_MESH, MECHA_FEM, IM_MECHA,NSTR_SUBDIV,MECHA_ORIGIN,MECHA_SIZES, MESH_FILE;
	size_type N_INC,FINITE_STRAIN;	
	plain_vector U;                   //solutions
	finite_strain_problem(std::string finput):
		read_flag(read_input(finput)), 
                med_mechanic(PARAM.string_value("MESH_FILE"), "Mesh_1"),   
                m_mechanic(med_mechanic.get_getfem_mesh()),
		mat_implant_param(2),
		mf_mechanic(m_mechanic),
		mim_mechanic(m_mechanic),
		mf_data(m_mechanic)
		{}
	bool read_input(std::string);
	void init(void);
	void solve(void);
	void export_deformed_VTK(getfem::mesh_fem& mf, plain_vector& def_U, plain_vector& res_U,std::string fname);
};
bool finite_strain_problem::read_input(std::string finput)
{
//Reading the parameter file
     std::cout<<"Reading input ......"<<std::endl;
     std::ifstream infile(finput.c_str());
     PARAM.read_param_file(infile);
     return true;
}
void finite_strain_problem::init()
{
     N_INC=PARAM.int_value("N_INC");
     TOLERANCE_NEWTON=PARAM.real_value("TOLERANCE_NEWTON");
     P_FINAL=PARAM.real_value("P_FINAL");
     MECHA_MESH=PARAM.string_value("MECHA_MESH");
     NSTR_SUBDIV=PARAM.string_value("NSTR_SUBDIV");
     MECHA_ORIGIN=PARAM.string_value("MECHA_ORIGIN");
     MECHA_SIZES=PARAM.string_value("MECHA_SIZES");
     MECHA_FEM=PARAM.string_value("MECHA_FEM");
     IM_MECHA=PARAM.string_value("IM_MECHA");
     MAT_LAMBDA=PARAM.real_value("MAT_LAMBDA");
     MAT_MU=PARAM.real_value("MAT_MU"); 
     FINITE_STRAIN=PARAM.int_value("FINITE_STRAIN"); //decide if you run small or finite strain problem
     MAX_ITER=PARAM.int_value("MAX_ITER"); 
//
     std::cout<<"++++++++++++++++++++ BEGINNING OF THE ANALYSIS++++++++++++++++++++++"<<std::endl;

//Mechanic FEM
     size_type mesh_order = 2;
     mf_mechanic.set_classical_finite_element(mesh_order);
     mf_mechanic.set_qdim(m_mechanic.dim());
//DATA FEM
     mf_data.set_classical_finite_element(mesh_order);
//Mechanic integration method
     mim_mechanic.set_integration_method(m_mechanic.convex_index(),mesh_order);
//Initialize the potential and the displacement fields
     gmm::resize(U,mf_mechanic.nb_dof());
     gmm::clear(U);

//   Initializing the model state variable for the structural brick
     getfem::mdbrick_abstract<> *pelastic_implant;
     if (FINITE_STRAIN)  
		{getfem::SaintVenant_Kirchhoff_hyperelastic_law *pmat_law=new getfem::SaintVenant_Kirchhoff_hyperelastic_law;
     		mat_implant_param[0]=MAT_LAMBDA; mat_implant_param[1]=MAT_MU;
		pelastic_implant=new getfem::mdbrick_nonlinear_elasticity<>(*pmat_law,mim_mechanic,mf_mechanic,mat_implant_param);}
     else
	       {getfem::mdbrick_isotropic_linearized_elasticity<> *p_lin_elastic=new getfem::mdbrick_isotropic_linearized_elasticity<>(mim_mechanic, mf_mechanic);
     		p_lin_elastic->lambda().set(MAT_LAMBDA);
     		p_lin_elastic->mu().set(MAT_MU);
		pelastic_implant=p_lin_elastic;}
     
     size_type EA_WE_region = med_mechanic.get_region_index("fixed");
     size_type SOUTH_region = med_mechanic.get_region_index("south");
     getfem::mdbrick_Dirichlet<> *pdirichlet_brick = new getfem::mdbrick_Dirichlet<>(*pelastic_implant,EA_WE_region,mf_mechanic);
     pdirichlet_brick->set_constraints_type(getfem::PENALIZED_CONSTRAINTS);
     plain_vector F(m_mechanic.dim()*mf_data.nb_dof());
     pstruct_brick= new getfem::mdbrick_source_term<>(*pdirichlet_brick,mf_data,F,SOUTH_region);

}


void finite_strain_problem::solve()
{
	size_type inc=0;
	scalar_type p_current=0.0, delta_p=P_FINAL/scalar_type(N_INC);
	std::string file_mechanic="";
	size_type ind_minY=0;
	std::string table_file_name=PARAM.string_value("TABLE_FILE");
	std::ofstream table_file(table_file_name.c_str());
	getfem::standard_model_state MS(*pstruct_brick);
	gmm::iteration iter = gmm::iteration(TOLERANCE_NEWTON, 10, MAX_ITER);
	std::cout<<"Entering the increment loop"<<std::endl;
	for (inc=0; inc<=N_INC; inc++){
		p_current=inc*delta_p;
		std::cout<<"	Increment -------------------- "<<inc<<std::endl;
		std::string str_inc=itoa(inc,10);
		plain_vector Current_Pressure(mf_data.nb_dof()*m_mechanic.dim());
		gmm::fill(gmm::sub_vector(Current_Pressure,gmm::sub_slice(1,mf_data.nb_dof(),m_mechanic.dim())),(-1.0)*p_current);
		pstruct_brick->source_term().set(Current_Pressure);
	        //Solve
     		getfem::standard_solve(MS, *pstruct_brick, iter);
		std::cout<<"Step converged ? - "<<(iter.converged()?"yeap":"NO!!")<<std::endl;
		gmm::copy(MS.state(),U);
		//
		//Postprocess the increment
		std::cout<<"	Writing the results"<<std::endl;
		file_mechanic=str_inc+"_structure";
		export_deformed_VTK(mf_mechanic, U, U,file_mechanic);
                med_mechanic.write_nodal_field(mf_mechanic,U,inc,scalar_type(inc),"displacement","disp.med");
		if (inc==0) {
			DVtable zero_point;
			zero_point.dispY=0.0;
			zero_point.volt=0.0;
			dvtab.resize(inc+1);
			dvtab[inc]=zero_point; 
		}
		scalar_type min=0.0;
		for (size_type i=0;i<U.size();i++) if (U[i]<min) {min=U[i]; ind_minY=i;}
		DVtable point;
		point.dispY=fabs(U[ind_minY]);
		point.volt=p_current;
		dvtab.resize(inc+1);
		dvtab[inc]=point;

		std::cout<<"	at p="<<dvtab[inc].volt<<"Pa,	deflection="<<dvtab[inc].dispY<<" ind_minY="<<ind_minY<<std::endl;
		table_file<<dvtab[inc].dispY<<" "<<dvtab[inc].volt<<std::endl;
	}
	table_file.close();
	std::cout<<"++++++++++++++++++++ END OF THE ANALYSIS+++++++++++++++++++++++++++"<<std::endl;
}

int main(int argc, char *argv[]) {


  DAL_SET_EXCEPTION_DEBUG; // Exceptions make a memory fault, to debug.
  FE_ENABLE_EXCEPT;        // Enable floating point exception for Nan.

  try {
    std::string finput(argv[1]);
    finite_strain_problem p(finput);
    p.init();
    p.solve();
    }
  catch(MEDMEM::MEDEXCEPTION& me) {std::cout<<me.what()<<std::endl;}
  DAL_STANDARD_CATCH_ERROR;
  return 0; 
}

void finite_strain_problem::export_deformed_VTK(getfem::mesh_fem& mf, plain_vector& def_U, plain_vector& res_U,std::string fname)
{	getfem::mesh* pmesh_deformed=deform_mesh(mf, def_U);
        getfem::mesh_fem mf_deformed(*pmesh_deformed);
	size_type Qsize = (gmm::vect_size(res_U)*mf.get_qdim() / mf.nb_dof()); 
	mf_deformed.set_qdim(Qsize);
        mf_deformed.set_finite_element(pmesh_deformed->convex_index(), mf.fem_of_element(0));
	std::string vtk_file_name=fname+".vtk";
        export_vtk(mf_deformed,res_U,vtk_file_name);
	delete pmesh_deformed;
}


void export_vtk(const getfem::mesh_fem& mf, const std::vector<bgeot::scalar_type>& U, std::string fname)
{    getfem::vtk_export exp(fname,false);
     exp.exporting(mf);
     exp.write_point_data(mf, U,"scalar");
}


getfem::mesh* deform_mesh(const getfem::mesh_fem& mf,plain_vector& def_U)
{	getfem::mesh* pmesh_deformed=new getfem::mesh();
	dal::bit_vector nn=mf.linked_mesh().convex_index(); 
        bgeot::size_type i; 
	for(i<<nn; i!=bgeot::size_type(-1); i<<nn) 
        {  
		getfem::mesh::ind_cv_ct point=mf.linked_mesh().ind_points_of_convex(i);
                getfem::mesh_fem::ind_dof_ct dof=mf.ind_basic_dof_of_element(i);
                bgeot::size_type num_points=mf.linked_mesh().structure_of_convex(i)->nb_points(); 
		std::vector<bgeot::base_node> pt(num_points);
        	for(bgeot::size_type j=0; j<num_points; j++) 
        		{ bgeot::base_node node=mf.linked_mesh().points()[point[j]]; 
			  if(mf.linked_mesh().dim()==2) {pt[j]=bgeot::base_node(node[0]+def_U[dof[2*j]],node[1]+def_U[dof[2*j+1]]);}
			  if(mf.linked_mesh().dim()==3) {pt[j]=bgeot::base_node(node[0]+def_U[dof[3*j]],node[1]+def_U[dof[3*j+1]],node[2]+def_U[dof[3*j+2]]);}
			}
        	pmesh_deformed->add_convex_by_points(mf.linked_mesh().trans_of_convex(i),pt.begin());
        }
        copy_regions(mf.linked_mesh(),*pmesh_deformed);
	return pmesh_deformed;
} 

void copy_regions(const getfem::mesh& from_mesh, getfem::mesh& to_mesh)
{
    for(dal::bv_visitor r(from_mesh.regions_index());!r.finished(); ++r)
    {
        for(getfem::mr_visitor i(from_mesh.region(r));!i.finished(); ++i)
        {
         if(i.is_face()) 
                to_mesh.region(r).add(i.cv(), i.f());
         else
                to_mesh.region(r).add(i.cv());
        }
    
    }

}
